@@ -1,3 +1,5 @@
+package edu.sc.seis.TauP;
+
 /*
  * The TauP Toolkit: Flexible Seismic Travel-Time and Raypath Utilities.
  * Copyright (C) 1998-2000 University of South Carolina This program is free
@@ -14,13 +16,14 @@
  * should be directed to H. Philip Crotwell, crotwell@seis.sc.edu or Tom Owens,
  * owens@seis.sc.edu
  */
-package edu.sc.seis.TauP;
 
 import java.io.DataInputStream;
 import java.io.DataOutputStream;
 import java.io.IOException;
 import java.io.Serializable;
 import java.lang.reflect.InvocationTargetException;
+import java.util.ArrayList;
+import java.util.List;
 
 /**
  * provides storage and methods for distance, time and tau increments for a
@@ -33,7 +36,11 @@ import java.lang.reflect.InvocationTargetException;
 public class TauBranch implements Serializable, Cloneable {
 
     /** Turns on debugging output. */
-    transient public boolean DEBUG = ToolRun.DEBUG;
+    transient public boolean DEBUG = TauPConfig.DEBUG;
+
+    protected SlownessModel sMod;
+    protected int topLayerNum;
+    protected int botLayerNum;
 
     /** The type of wave for this branch, P or S. */
     protected boolean isPWave;
@@ -77,8 +84,15 @@ public class TauBranch implements Serializable, Cloneable {
     /** Holds tau evaluated at the ith ray parameter for this branch. */
     protected double[] tau;
 
+    /** Slowness at top of top layer of branch. */
+    private double topRayParam;
+
+    /** Slowness at bottom of bottom layer of branch. */
+    private double botRayParam;
+
     // Constructors --------------------------------------------------------
-    public TauBranch(boolean isPWave,
+    public TauBranch(SlownessModel sMod,
+                     boolean isPWave,
                      double topDepth,
                      double botDepth,
                      double maxRayParam,
@@ -87,6 +101,7 @@ public class TauBranch implements Serializable, Cloneable {
                      double[] dist,
                      double[] time,
                      double[] tau) {
+        this.sMod = sMod;
         this.isPWave = isPWave;
         this.topDepth = topDepth;
         this.botDepth = botDepth;
@@ -98,11 +113,21 @@ public class TauBranch implements Serializable, Cloneable {
         this.tau = tau;
     }
     
-    public TauBranch(double topDepth, double botDepth, boolean isPWave) {
+    public TauBranch(SlownessModel sMod,double topDepth, double botDepth, boolean isPWave) throws NoSuchLayerException {
+        this.sMod = sMod;
         this.topDepth = topDepth;
         this.botDepth = botDepth;
         this.isPWave = isPWave;
-        }
+
+        this.topLayerNum = sMod.layerNumberBelow(getTopDepth(), isPWave);
+        this.botLayerNum = sMod.layerNumberAbove(getBotDepth(), isPWave);
+        SlownessLayer topSLayer = sMod.getSlownessLayer(topLayerNum,
+                isPWave);
+        SlownessLayer botSLayer = sMod.getSlownessLayer(botLayerNum,
+                isPWave);
+        this.topRayParam = topSLayer.getTopP();
+        this.botRayParam = botSLayer.getBotP();
+    }
 
     // Methods -------------------------------------------------------------
     // Accessor methods
@@ -116,6 +141,16 @@ public class TauBranch implements Serializable, Cloneable {
         return botDepth;
     }
 
+    /** Slowness at top of top layer of branch. */
+    public double getTopRayParam() {
+        return topRayParam;
+    }
+
+    /** Slowness at bottom of bottom layer of branch. */
+    public double getBotRayParam() {
+        return botRayParam;
+    }
+
     /**
      * @return the maximum ray parameter that can penetrate into this branch.
      *          Time, distance, and tau increments are undefined, set to 0.0,
@@ -146,7 +181,7 @@ public class TauBranch implements Serializable, Cloneable {
      *          i_th ray parameter for this branch.
      */
     public double[] getDist() {
-        return (double[])dist.clone();
+        return dist.clone();
     }
 
     /**
@@ -162,7 +197,7 @@ public class TauBranch implements Serializable, Cloneable {
      *          i_th ray parameter for this branch.
      */
     public double[] getTime() {
-        return (double[])time.clone();
+        return time.clone();
     }
 
     /**
@@ -178,7 +213,7 @@ public class TauBranch implements Serializable, Cloneable {
      *          ray parameter for this branch.
      */
     public double[] getTau() {
-        return (double[])tau.clone();
+        return tau.clone();
     }
 
     /**
@@ -201,33 +236,34 @@ public class TauBranch implements Serializable, Cloneable {
      * @exception TauModelException
      *                if the slownessmodel and taumodel are not compatible
      */
-    public void createBranch(SlownessModel sMod,
-                             double rayParams[]) throws NoSuchLayerException,
+    public static TauBranch createBranch(SlownessModel sMod,
+                             double[] rayParams,
+                             double topDepth, double botDepth, boolean isPWave, boolean DEBUG) throws
             SlownessModelException, TauModelException {
+        TauBranch tb = new TauBranch(sMod, topDepth, botDepth, isPWave);
+        tb.DEBUG = DEBUG;
         TimeDist timeDist;
         double p;
-        int topLayerNum = sMod.layerNumberBelow(getTopDepth(), isPWave);
-        int botLayerNum = sMod.layerNumberAbove(getBotDepth(), isPWave);
-        SlownessLayer topSLayer = sMod.getSlownessLayer(topLayerNum,
+        SlownessLayer topSLayer = sMod.getSlownessLayer(tb.topLayerNum,
                                                              isPWave);
-        SlownessLayer botSLayer = sMod.getSlownessLayer(botLayerNum,
+        SlownessLayer botSLayer = sMod.getSlownessLayer(tb.botLayerNum,
                                                              isPWave);
-        if(topSLayer.getTopDepth() != getTopDepth()
-                || botSLayer.getBotDepth() != getBotDepth()) {
-            if(topSLayer.getTopDepth() != getTopDepth()
-                    && Math.abs(topSLayer.getTopDepth() - getTopDepth()) < 0.000001) {
+        if(topSLayer.getTopDepth() != tb.getTopDepth()
+                || botSLayer.getBotDepth() != tb.getBotDepth()) {
+            if(topSLayer.getTopDepth() != tb.getTopDepth()
+                    && Math.abs(topSLayer.getTopDepth() - tb.getTopDepth()) < 0.000001) {
                 // really close, so just move top
-                System.err.println("Changing topDepth: " + "\ntopDepth: "
-                        + getTopDepth() + " " + topSLayer.getTopDepth()
-                        + "\nbotDepth: " + getBotDepth() + " "
+                Alert.warning("Changing topDepth: " + "\ntopDepth: "
+                        + tb.getTopDepth() + " " + topSLayer.getTopDepth()
+                        + "\nbotDepth: " + tb.getBotDepth() + " "
                         + botSLayer.getBotDepth());
                 topDepth = topSLayer.getTopDepth();
-            } else if(botSLayer.getBotDepth() != getBotDepth()
-                    && Math.abs(botSLayer.getBotDepth() - getBotDepth()) < 0.000001) {
+            } else if(botSLayer.getBotDepth() != tb.getBotDepth()
+                    && Math.abs(botSLayer.getBotDepth() - tb.getBotDepth()) < 0.000001) {
                 // really close, so just move bottom
-                System.err.println("Changing botDepth: " + "\ntopDepth: "
-                        + getTopDepth() + " " + topSLayer.getTopDepth()
-                        + "\nbotDepth: " + getBotDepth() + " "
+                Alert.warning("Changing botDepth: " + "\ntopDepth: "
+                        + tb.getTopDepth() + " " + topSLayer.getTopDepth()
+                        + "\nbotDepth: " + tb.getBotDepth() + " "
                         + botSLayer.getBotDepth());
                 botDepth = botSLayer.getBotDepth();
             } else {
@@ -236,11 +272,11 @@ public class TauBranch implements Serializable, Cloneable {
                         + "SP".charAt(isPWave ? 1 : 0)
                         + ":"
                         + "\ntopDepth: "
-                        + getTopDepth()
+                        + tb.getTopDepth()
                         + " "
                         + topSLayer.getTopDepth()
                         + "\nbotDepth: "
-                        + getBotDepth()
+                        + tb.getBotDepth()
                         + " "
                         + botSLayer.getBotDepth());
             }
@@ -254,71 +290,104 @@ public class TauBranch implements Serializable, Cloneable {
          */
         // max ray param cannot be larger than the largest of the top and bottom p
         // branches should be monotonic wrt slowness
-        maxRayParam = Math.max(topSLayer.getTopP(), botSLayer.getBotP());
-        minTurnRayParam = sMod.getMinTurnRayParam(getBotDepth(), isPWave);
-        minRayParam = sMod.getMinRayParam(getBotDepth(), isPWave);
-        tau = new double[rayParams.length];
-        dist = new double[rayParams.length];
-        time = new double[rayParams.length];
+        tb.maxRayParam = Math.max(topSLayer.getTopP(), botSLayer.getBotP());
+        tb.minTurnRayParam = sMod.getMinTurnRayParam(tb.getTopDepth(), tb.getBotDepth(), isPWave);
+        tb.minRayParam = sMod.getMinRayParam(tb.getBotDepth(), isPWave);
+
+        tb.tau = new double[rayParams.length];
+        tb.dist = new double[rayParams.length];
+        tb.time = new double[rayParams.length];
 
         for(int rayNum = 0; rayNum < rayParams.length; rayNum++) {
             p = rayParams[rayNum];
-            timeDist = calcTimeDist(sMod, topLayerNum, botLayerNum, p, false);
-            dist[rayNum] = timeDist.getDistRadian();
-            time[rayNum] = timeDist.getTime();
-            tau[rayNum] = time[rayNum] - p * dist[rayNum];
+            timeDist = tb.calcTimeDist(p, false);
+            tb.dist[rayNum] = timeDist.getDistRadian();
+            tb.time[rayNum] = timeDist.getTime();
+            tb.tau[rayNum] = tb.time[rayNum] - p * tb.dist[rayNum];
         }
+        return tb;
     }
 
     /**
      * calculates the time and distance increments for the given ray parameter.
      * The topDepth and botDepth must be correct as they determine the bounds on
      * the integration/summing.
-     * 
-     * @exception NoSuchLayerException
-     *                if topLayerNum or botLayerNum are not in the slowness
-     *                model.
+     *
      * @exception SlownessModelException
      *                if the ray with ray parameter p turns within a layer
      *                instead of at the bottom.
      */
-    public TimeDist calcTimeDist(SlownessModel sMod,
-                                 int topLayerNum,
-                                 int botLayerNum,
-                                 double p) throws NoSuchLayerException,
+    public TimeDist calcTimeDist(double p) throws
             SlownessModelException {
-        return calcTimeDist(sMod, topLayerNum, botLayerNum, p, false);
+        return calcTimeDist(p, false);
     }
 
-    public TimeDist calcTimeDist(SlownessModel sMod,
-                                 int topLayerNum,
-                                 int botLayerNum,
-                                 double p,
-                                 boolean allowTurnInLayer) throws NoSuchLayerException,
+    public TimeDist calcTimeDist(double p,
+                                 boolean allowTurnInLayer) throws
             SlownessModelException {
+        return calcTimeDist(p, allowTurnInLayer, true);
+    }
+    public TimeDist calcTimeDist(double p,
+                                 boolean allowTurnInLayer,
+                                 boolean isDowngoing) throws SlownessModelException {
         int layerNum;
         TimeDist timeDist = new TimeDist(p);
-        SlownessLayer layer;
-        if(p <= getMaxRayParam()) {
-            layerNum = topLayerNum;
-            layer = sMod.getSlownessLayer(layerNum, isPWave);
-            while(layerNum <= botLayerNum && p <= layer.getTopP()
-                    && p <= layer.getBotP()) {
-                timeDist = timeDist.add(sMod.layerTimeDist(p, layerNum, isPWave));
-                layerNum++;
-                if(layerNum <= botLayerNum) {
-                    layer = sMod.getSlownessLayer(layerNum, isPWave);
+        if (isDowngoing) {
+            if (p <= getMaxRayParam()) {
+                // calculation goes top to bottom, even
+                layerNum = topLayerNum;
+                while (layerNum <= botLayerNum) {
+                    SlownessLayer layer = sMod.getSlownessLayer(layerNum, isPWave);
+                    if (p > layer.getTopP()) {
+                        // can't enter layer from top
+                        break;
+                    } else if (p <= layer.getTopP()
+                            && p <= layer.getBotP()) {
+                        // can cross layer
+                        timeDist = timeDist.add(sMod.layerTimeDist(p, layerNum, isPWave, isDowngoing));
+                    } else if (p < layer.getTopP()
+                            && p > layer.getBotP()) {
+                        // turns in layer
+                        if (allowTurnInLayer) {
+                            timeDist = timeDist.add(sMod.layerTimeDistAllowTurn(p, layerNum, isPWave, isDowngoing));
+                        } else {
+                            throw new SlownessModelException("Ray turns in the middle of this"
+                                    + " layer. layerNum = " + layerNum + (isPWave ? " P" : " S")
+                                    + (isDowngoing ? " down " : " up ")
+                                    + " sphericalRayParam " + p + " layer =" + layer);
+                        }
+                        break;
+                    }
+                    layerNum++;
                 }
             }
-            if((layer.getTopP() - p) * (p - layer.getBotP()) > 0) {
-                if (allowTurnInLayer) {
-                    timeDist = timeDist.add(sMod.layerTimeDistAllowTurn(p, layerNum, isPWave, true));
-                } else {
-                    throw new SlownessModelException("Ray turns in the middle of this"
-                            + " layer. layerNum = "
-                            + layerNum
-                            + (isPWave ? " P" : " S")
-                            + " sphericalRayParam " + p + " layer =" + layer);
+        } else {
+            // upgoing
+            if (p <= getMaxRayParam()) {
+                layerNum = botLayerNum;
+                while (layerNum >= topLayerNum) {
+                    SlownessLayer layer = sMod.getSlownessLayer(layerNum, isPWave);
+                    if (p > layer.getBotP()) {
+                        // can't enter layer from bot
+                        break;
+                    } else if (p <= layer.getTopP()
+                            && p <= layer.getBotP()) {
+                        // can cross layer
+                        timeDist = timeDist.add(sMod.layerTimeDist(p, layerNum, isPWave, isDowngoing));
+                    } else if (p < layer.getBotP()
+                            && p > layer.getTopP()) {
+                        // turns in layer
+                        if (allowTurnInLayer) {
+                            timeDist = timeDist.add(sMod.layerTimeDistAllowTurn(p, layerNum, isPWave, isDowngoing));
+                        } else {
+                            throw new SlownessModelException("Ray turns in the middle of this"
+                                    + " layer. layerNum = " + layerNum + (isPWave ? " P" : " S")
+                                    + (isDowngoing ? " down " : " up ")
+                                    + " sphericalRayParam " + p + " layer =" + layer);
+                        }
+                        break;
+                    }
+                    layerNum--;
                 }
             }
         }
@@ -336,7 +405,7 @@ public class TauBranch implements Serializable, Cloneable {
      * see edu.sc.seis.TauP.TauModel.depthCorrect(double)
      */
     protected void insert(double rayParam, SlownessModel sMod, int index)
-            throws NoSuchLayerException, SlownessModelException,
+            throws SlownessModelException,
             TauModelException {
         int topLayerNum = sMod.layerNumberBelow(getTopDepth(), isPWave);
         int botLayerNum = sMod.layerNumberAbove(getBotDepth(), isPWave);
@@ -378,7 +447,7 @@ public class TauBranch implements Serializable, Cloneable {
      * generates a new tau branch by "subtracting" the given tau branch from
      * this tau branch. The given tau branch is assumed to by the upper part of
      * this branch.
-     * 
+     * <p>
      *  indexP specifies where a new ray coresponding to a P wave
      *           sample has been added, it is -1 if no ray parameter has been
      *           added to topBranch.
@@ -391,8 +460,8 @@ public class TauBranch implements Serializable, Cloneable {
                                    int indexP,
                                    int indexS,
                                    SlownessModel sMod,
-                                   double rayParams[])
-            throws NoSuchLayerException, SlownessModelException,
+                                   double[] rayParams)
+            throws SlownessModelException,
             TauModelException {
         if(topBranch.getTopDepth() != getTopDepth()
                 || topBranch.getBotDepth() > getBotDepth()) {
@@ -478,12 +547,12 @@ public class TauBranch implements Serializable, Cloneable {
                     + rayParams[indexS]
                     + " " + rayParams[indexS + 1] + "\n" + sLayer);
         }
-        sLayer = null;
         // construct the new TauBranch, going from the bottom of
         // the top half to the bottom of the whole branch
-        TauBranch botBranch = new TauBranch(topBranch.getBotDepth(),
+        TauBranch botBranch = new TauBranch(sMod, topBranch.getBotDepth(),
                                             getBotDepth(),
                                             isPWave);
+        botBranch.sMod = sMod;
         botBranch.maxRayParam = topBranch.getMinRayParam();
         botBranch.minTurnRayParam = getMinTurnRayParam();
         botBranch.minRayParam = getMinRayParam();
@@ -494,18 +563,12 @@ public class TauBranch implements Serializable, Cloneable {
         if(indexP != -1) {
             arrayLength++;
             PRayParam = rayParams[indexP];
-            timeDistP = botBranch.calcTimeDist(sMod,
-                                               topLayerNum,
-                                               botLayerNum,
-                                               PRayParam);
+            timeDistP = botBranch.calcTimeDist(PRayParam);
         }
         if(indexS != -1 && indexS != indexP) {
             arrayLength++;
             SRayParam = rayParams[indexS];
-            timeDistS = botBranch.calcTimeDist(sMod,
-                                               topLayerNum,
-                                               botLayerNum,
-                                               SRayParam);
+            timeDistS = botBranch.calcTimeDist(SRayParam);
         } else {
             // in case indexS==indexP then we only need one
             indexS = -1;
@@ -586,13 +649,21 @@ public class TauBranch implements Serializable, Cloneable {
         tau = newTau;
     }
 
-    public TimeDist[] path(double rayParam,
-                           boolean downgoing,
-                           SlownessModel sMod) throws SlownessModelException {
+
+    public List<TimeDist> path(double rayParam,
+                               boolean downgoing,
+                               SlownessModel sMod) throws SlownessModelException {
+        return path(rayParam, downgoing, sMod, new TimeDist(rayParam));
+    }
+    public List<TimeDist> path(double rayParam,
+                               boolean downgoing,
+                               SlownessModel sMod,
+                               TimeDist prevTD) throws SlownessModelException {
+        List<TimeDist> thePath = new ArrayList<>();
         if(rayParam > getMaxRayParam()) {
-            return null;
+            return thePath;
         }
-        Assert.isTrue(rayParam >= 0.0, "ray parameter must not be negative.");
+        if (rayParam < 0.0) {throw new IllegalArgumentException("ray parameter must not be negative.");}
         int topLayerNum;
         int botLayerNum;
         try {
@@ -603,12 +674,10 @@ public class TauBranch implements Serializable, Cloneable {
                     + "SlownessModel and TauModel are out of sync. "
                     + e.getMessage());
         }
-        TimeDist[] thePath = new TimeDist[botLayerNum - topLayerNum + 2];
         int sLayerNum;
-        int pathIndex = 0;
         double turnDepth;
         SlownessLayer sLayer, turnSLayer;
-        /** check to make sure layers and branches are compatable. */
+        /* check to make sure layers and branches are compatable. */
         sLayer = sMod.getSlownessLayer(topLayerNum, isPWave);
         if(sLayer.getTopDepth() != getTopDepth()) {
             throw new SlownessModelException("Branch and Slowness model are not compatible! "
@@ -628,15 +697,15 @@ public class TauBranch implements Serializable, Cloneable {
         if(downgoing) {
             sLayerNum = topLayerNum;
             sLayer = sMod.getSlownessLayer(sLayerNum, isPWave);
-            thePath[pathIndex] = new TimeDist(rayParam, 0, 0, sLayer.getTopDepth());
-            pathIndex++;
+            prevTD = prevTD.add(new TimeDist(rayParam, 0, 0, sLayer.getTopDepth()));
+            thePath.add( prevTD);
             while(sLayer.getBotP() >= rayParam && sLayerNum <= botLayerNum) {
                 if(!sLayer.isZeroThickness()) {
-                    thePath[pathIndex] = sMod.layerTimeDist(rayParam,
+                    prevTD = prevTD.add( sMod.layerTimeDist(rayParam,
                                                             sLayerNum,
                                                             isPWave,
-                                                            downgoing);
-                    pathIndex++;
+                                                            downgoing));
+                    thePath.add( prevTD);
                 }
                 sLayerNum++;
                 if(sLayerNum <= botLayerNum) {
@@ -650,10 +719,10 @@ public class TauBranch implements Serializable, Cloneable {
                                                sLayer.getTopDepth(),
                                                rayParam,
                                                turnDepth);
-                thePath[pathIndex] = turnSLayer.bullenRadialSlowness(rayParam,
+                prevTD = prevTD.add( turnSLayer.bullenRadialSlowness(rayParam,
                                                                      sMod.getRadiusOfEarth(),
-                                                                     downgoing);
-                pathIndex++;
+                                                                     downgoing));
+                thePath.add( prevTD);
             }
         } else {
             // upgoing
@@ -667,32 +736,32 @@ public class TauBranch implements Serializable, Cloneable {
                 // turned in layer, so initial is turn point
                 turnDepth = sLayer.bullenDepthFor(rayParam,
                                                   sMod.getRadiusOfEarth());
-                thePath[pathIndex] = new TimeDist(rayParam, 0, 0, turnDepth);
-                pathIndex++;
+                prevTD = prevTD.add( new TimeDist(rayParam, 0, 0, turnDepth));
+                thePath.add( prevTD);
                 turnSLayer = new SlownessLayer(sLayer.getTopP(),
                                                sLayer.getTopDepth(),
                                                rayParam,
                                                turnDepth);
-                thePath[pathIndex] = turnSLayer.bullenRadialSlowness(rayParam,
+                prevTD = prevTD.add(turnSLayer.bullenRadialSlowness(rayParam,
                                                                      sMod.getRadiusOfEarth(),
-                                                                     downgoing);
-                pathIndex++;
+                                                                     downgoing));
+                thePath.add( prevTD);
                 sLayerNum--;
                 if(sLayerNum >= topLayerNum) {
                     sLayer = sMod.getSlownessLayer(sLayerNum, isPWave);
                 }
             } else {
                 // enter from bottom,
-                thePath[pathIndex] = new TimeDist(rayParam, 0, 0, sLayer.getBotDepth());
-                pathIndex++;
+                prevTD = prevTD.add(new TimeDist(rayParam, 0, 0, sLayer.getBotDepth()));
+                thePath.add( prevTD);
             }
             while(sLayerNum >= topLayerNum) {
                 if(!sLayer.isZeroThickness()) {
-                    thePath[pathIndex] = sMod.layerTimeDist(rayParam,
+                    prevTD = prevTD.add(sMod.layerTimeDist(rayParam,
                                                             sLayerNum,
                                                             isPWave,
-                                                            downgoing);
-                    pathIndex++;
+                                                            downgoing));
+                    thePath.add( prevTD);
                 }
                 sLayerNum--;
                 if(sLayerNum >= topLayerNum) {
@@ -700,9 +769,7 @@ public class TauBranch implements Serializable, Cloneable {
                 }
             }
         }
-        TimeDist[] tempPath = new TimeDist[pathIndex];
-        System.arraycopy(thePath, 0, tempPath, 0, pathIndex);
-        return tempPath;
+        return thePath;
     }
 
     public void writeToStream(DataOutputStream dos) throws IOException {
@@ -734,12 +801,10 @@ public class TauBranch implements Serializable, Cloneable {
         byte[] classString = new byte[dis.readInt()];
         dis.read(classString);
         Class tBranchClass = Class.forName(new String(classString));
-        TauBranch tBranch = null;
+        TauBranch tBranch;
         try {
             tBranch = (TauBranch)tBranchClass.getDeclaredConstructor().newInstance();
-        } catch (InvocationTargetException e) {
-            throw new TauPException("Problem creating TauBranch", e);
-        } catch (NoSuchMethodException e) {
+        } catch (InvocationTargetException | NoSuchMethodException e) {
             throw new TauPException("Problem creating TauBranch", e);
         }
         tBranch.topDepth = dis.readDouble();
@@ -779,26 +844,42 @@ public class TauBranch implements Serializable, Cloneable {
 //        System.arraycopy(time, 0, newTime, 0, dist.length);
 //        double[] newTau = new double[tau.length];
 //        System.arraycopy(tau, 0, newTau, 0, dist.length);
-        return new TauBranch(isPWave,
-                             topDepth,
-                             botDepth,
-                             maxRayParam,
-                             minTurnRayParam,
-                             minRayParam,
-                             dist, time, tau);
+        TauBranch tb = new TauBranch(sMod,
+                isPWave,
+                topDepth,
+                botDepth,
+                maxRayParam,
+                minTurnRayParam,
+                minRayParam,
+                dist, time, tau);
+        tb.sMod = sMod;
+        tb.topLayerNum = topLayerNum;
+        tb.botLayerNum = botLayerNum;
+        tb.topRayParam = topRayParam;
+        tb.botRayParam = botRayParam;
+        return tb;
     }
 
     public String toString() {
-        String desc = "Tau Branch\n";
+        String desc = "Tau Branch "+(isPWave?"P":"S")+" "+"\n";
         desc += " topDepth = " + getTopDepth() + "\n";
         desc += " botDepth = " + getBotDepth() + "\n";
         desc += " maxRayParam=" + getMaxRayParam() + " minTurnRayParam="
                 + getMinTurnRayParam();
         desc += " minRayParam=" + getMinRayParam() + "\n";
+        desc += " topRP="+getTopRayParam()+" botRP="+getBotRayParam();
         /*
          * for (int i=0;i <tau.length;i++) { desc += "i = "+i+" time="+time[i]+"
          * dist="+dist[i]+" tau="+tau[i]+"\n"; }
          */
         return desc;
     }
+
+    public boolean isHighSlowness() {
+        SlownessLayer topSLayer = sMod.getSlownessLayer(topLayerNum,
+                isPWave);
+        SlownessLayer botSLayer = sMod.getSlownessLayer(botLayerNum,
+                isPWave);
+        return topSLayer.getTopP() <= botSLayer.getBotP();
+    }
 }
