@@ -16,19 +16,7 @@
  */
 package edu.sc.seis.TauP;
 
-import java.io.BufferedInputStream;
-import java.io.FileInputStream;
-import java.io.FileNotFoundException;
-import java.io.FileOutputStream;
-import java.io.IOException;
-import java.io.InputStream;
-import java.io.InvalidClassException;
-import java.io.ObjectInputStream;
-import java.io.ObjectOutputStream;
-import java.io.OptionalDataException;
-import java.io.OutputStream;
-import java.io.Serializable;
-import java.io.StreamCorruptedException;
+import java.io.*;
 import java.lang.ref.SoftReference;
 import java.util.Arrays;
 import java.util.HashMap;
@@ -41,9 +29,10 @@ import java.util.HashMap;
  */
 public class TauModel implements Serializable {
 
-    public TauModel(SlownessModel sMod) throws NoSuchLayerException,
-            NoSuchMatPropException, SlownessModelException, TauModelException {
-        this.sMod = (SlownessModel)sMod;
+
+    public TauModel(SlownessModel sMod) throws
+            SlownessModelException, TauModelException {
+        this.sMod = sMod;
         calcTauIncFrom();
     }
 
@@ -78,7 +67,7 @@ public class TauModel implements Serializable {
     }
 
     /** True to enable debugging output. */
-    transient public static boolean DEBUG = ToolRun.DEBUG;
+    public static boolean DEBUG = TauPConfig.DEBUG;
 
     /** True if this is a spherical slowness model. False if flat. */
     protected boolean spherical = true;
@@ -98,6 +87,8 @@ public class TauModel implements Serializable {
      */
     protected double[] noDisconDepths = new double[0];
 
+    protected double surfaceDepth = 0;
+
     /** Depth of the moho. */
     protected double mohoDepth;
 
@@ -125,7 +116,7 @@ public class TauModel implements Serializable {
      * an event at depth. This is normally be set when the tau model is
      * generated to be a clone of the slowness model.
      */
-    private SlownessModel sMod;
+    private final SlownessModel sMod;
 
     /**
      * ray parameters used to construct the tau branches. This may only be a
@@ -194,8 +185,86 @@ public class TauModel implements Serializable {
         return false;
     }
 
+    public boolean isFluidBranch(int branchNum) {
+        return getSlownessModel().depthInFluid(getTauBranch(branchNum, false).getTopDepth());
+    }
+
+    /**
+     * True if a boundary can generate a head wave, must be a discontinuity and an increase in velocity with depth.
+     *
+     * @param branchNum branch layer number
+     * @param isPWave true for P, false for S
+     * @return head wave possible
+     * @throws NoSuchLayerException
+     */
+    public boolean isHeadWaveBranch(int branchNum, boolean isPWave) throws NoSuchLayerException {
+        if (getTauBranch(branchNum, isPWave).isHighSlowness()) {
+            return false;
+        }
+        double topDepth = getTauBranch(branchNum, false).getTopDepth();
+        int aboveIdx = getVelocityModel().layerNumberAbove(topDepth);
+        VelocityLayer above = getVelocityModel().getVelocityLayer(aboveIdx);
+        int belowIdx = getVelocityModel().layerNumberBelow(topDepth);
+        VelocityLayer below = getVelocityModel().getVelocityLayer(belowIdx);
+        if (isPWave) {
+            return above.getBotPVelocity() < below.getTopPVelocity();
+        } else {
+            return above.getBotSVelocity() < below.getTopSVelocity();
+        }
+    }
+
+    /**
+     * True if a boundary can generate a diffracted wave, currently just ensure a discontinuity.
+     *
+     * @param branchNum branch layer number
+     * @param isPWave true for P, false for S
+     * @return diffracted wave possible
+     */
+    public boolean isDiffractionBranch(int branchNum, boolean isPWave) {
+        return isDiscontinuityBranch(branchNum, isPWave);
+    }
+
+    /**
+     * True if a boundary at top of branch is a discontinuity for the given phase type in the velocity model.
+     * Branch 0 is always a discontinuity, free surface. Center of earth is never a discon, so we do not need
+     * to search the bottom of a branch.
+     *
+     * @param branchNum branch layer number
+     * @param isPWave true for P, false for S
+     * @return if a velocity discontinuity
+     */
+    public boolean isDiscontinuityBranch(int branchNum, boolean isPWave) {
+        if (branchNum==0) {
+            // free surface is always a discon
+            return true;
+        }
+        double topDepth = getTauBranch(branchNum, isPWave).getTopDepth();
+        int aboveIdx;
+        try {
+            aboveIdx = getVelocityModel().layerNumberAbove(topDepth);
+        } catch (NoSuchLayerException e) {
+            // no above means free surface, so is discon
+            return true;
+        }
+        VelocityLayer above = getVelocityModel().getVelocityLayer(aboveIdx);
+        int belowIdx;
+        try {
+            belowIdx = getVelocityModel().layerNumberBelow(topDepth);
+        } catch (NoSuchLayerException e) {
+            // no layer below means center of earth, so not discon
+            return false;
+        }
+        VelocityLayer below = getVelocityModel().getVelocityLayer(belowIdx);
+        if (isPWave) {
+            return above.getBotPVelocity() != below.getTopPVelocity();
+        } else {
+            return above.getBotSVelocity() != below.getTopSVelocity();
+        }
+    }
+
     /**
      * Is the given depth a "noDisconDepth"?
+     * Usually because model was split at the source or receiver depth, or an added depth used by taup pierce.
      */
     public boolean isNoDisconDepth(double noDisconDepth) {
         for(int i = 0; i < noDisconDepths.length; i++) {
@@ -227,6 +296,16 @@ public class TauModel implements Serializable {
         return mohoBranch;
     }
 
+    /** @return tau branch with the moho at its top. */
+    public TauBranch getAboveMohoTauBranch(boolean isPWave) {
+        return getTauBranch(mohoBranch-1, isPWave);
+    }
+
+    /** @return tau branch with the moho at its top. */
+    public TauBranch getBelowMohoTauBranch(boolean isPWave) {
+        return getTauBranch(mohoBranch, isPWave);
+    }
+
     /** @return depth of the cmb. */
     public double getCmbDepth() {
         return cmbDepth;
@@ -237,6 +316,16 @@ public class TauModel implements Serializable {
         return cmbBranch;
     }
 
+    /** @return tau branch with the iocb at its top. */
+    public TauBranch getAboveCmbTauBranch(boolean isPWave) {
+        return getTauBranch(cmbBranch-1, isPWave);
+    }
+
+    /** @return tau branch with the iocb at its top. */
+    public TauBranch getBelowCmbTauBranch(boolean isPWave) {
+        return getTauBranch(cmbBranch, isPWave);
+    }
+
     /** @return depth of the iocb. */
     public double getIocbDepth() {
         return iocbDepth;
@@ -247,6 +336,52 @@ public class TauModel implements Serializable {
         return iocbBranch;
     }
 
+    /** @return tau branch with the iocb at its bottom. */
+    public TauBranch getAboveIocbTauBranch(boolean isPWave) {
+        return getTauBranch(iocbBranch-1, isPWave);
+    }
+
+    /** @return tau branch with the iocb at its top. */
+    public TauBranch getBelowIocbTauBranch(boolean isPWave) {
+        return getTauBranch(iocbBranch, isPWave);
+    }
+
+    /**
+     * Check if crust is missing, moho at surface
+     * @return
+     */
+    public boolean isDegenerateCrust() {
+        // degenerate case, moho at surface
+        return getMohoDepth() == 0;
+    }
+
+    /**
+     * Check if mantle is missing, model is only core, cmb at surface
+     * @return
+     */
+    public boolean isDegenerateCrustMantle() {
+        // degenerate case, cmb is surface
+        return getCmbDepth() == 0;
+    }
+
+    /**
+     * Check if inner core is missing, iocb at center of earth
+     * @return
+     */
+    public boolean isDegenerateInnerCore() {
+        return getIocbDepth() == getRadiusOfEarth() ;
+        // degenerate case, IOCB is at center, so model without an inner core
+    }
+
+    /**
+     * Check if outer core is missing, iocb at cmb
+     * @return
+     */
+    public boolean isDegenerateOuterCore() {
+        return getIocbDepth() == getCmbDepth() ;
+        // degenerate case, IOCB is at CMB, so model without an outer core
+    }
+
     /**
      * @return the radius of the Earth in km, usually input from the velocity
      *          model.
@@ -262,7 +397,7 @@ public class TauModel implements Serializable {
      *          zones (low velocity zones).
      */
     public double[] getRayParams() {
-        return (double[])rayParams.clone();
+        return rayParams.clone();
     }
 
     public double getOneRayParam(int i) {
@@ -273,6 +408,18 @@ public class TauModel implements Serializable {
         return tauBranches[0].length;
     }
 
+
+    /**
+     * Gets the branch that either has the depth as its top boundary, or
+     * strictly contains the depth. Also, we allow the bottommost branch to
+     * contain its bottom depth, so that the center if the earth is contained
+     * within the bottom branch.
+     */
+    public TauBranch getTauBranchAtDepth(double depth, boolean isPWave) throws TauModelException {
+        int idx = findBranch(depth);
+        return getTauBranch(idx, isPWave);
+    }
+
     public TauBranch getTauBranch(int branchNum, boolean isPWave) {
         if(isPWave) {
             return tauBranches[0][branchNum];
@@ -321,7 +468,7 @@ public class TauModel implements Serializable {
      *                distance without a layer.
      */
     private void calcTauIncFrom() throws SlownessModelException,
-            NoSuchLayerException, TauModelException, NoSuchMatPropException {
+            TauModelException {
         SlownessLayer topSLayer, botSLayer, currSLayer, prevSLayer;
         int topCritLayerNum, botCritLayerNum;
         /*
@@ -329,7 +476,7 @@ public class TauModel implements Serializable {
          * distance. Otherwise we must signal an exception.
          */
         if(DEBUG) {
-            System.out.println("Size of slowness model:"
+            Alert.debug("Size of slowness model:"
                     + " sMod.getNumLayers('P') = " + sMod.getNumLayers(true)
                     + ", sMod.getNumLayers('S') = " + sMod.getNumLayers(false));
         }
@@ -386,7 +533,6 @@ public class TauModel implements Serializable {
         double[] temptemprayParams = new double[rayNum];
         System.arraycopy(tempRayParams, 0, temptemprayParams, 0, rayNum);
         tempRayParams = temptemprayParams;
-        temptemprayParams = null;
         // sort
         Arrays.sort(tempRayParams);
         // and remove duplicates
@@ -406,7 +552,6 @@ public class TauModel implements Serializable {
                 n++;
             }
         }
-        tempRayParams = null;
         // reverse sort so large to small
         for (int i = 0; i < rayParams.length/2; i++) {
             double tmp = rayParams[i];
@@ -414,7 +559,7 @@ public class TauModel implements Serializable {
             rayParams[rayParams.length-i-1] = tmp;
         }
         if(DEBUG) {
-            System.out.println("Number of slowness samples for tau =" + numUnique);
+            Alert.debug("Number of slowness samples for tau =" + numUnique);
         }
         CriticalDepth topCritDepth, botCritDepth;
         int waveNum;
@@ -427,17 +572,16 @@ public class TauModel implements Serializable {
                 botCritDepth = sMod.getCriticalDepth(critNum + 1);
                 botCritLayerNum = botCritDepth.getLayerNum(isPWave) - 1;
                 if(DEBUG) {
-                    System.out.println("Calculating " + (isPWave ? "P" : "S")
+                    Alert.debug("Calculating " + (isPWave ? "P" : "S")
                             + " tau branch for branch " + critNum
                             + " topCritLayerNum=" + topCritLayerNum+" ("+topCritDepth.getDepth()+")"
                             + " botCritLayerNum=" + botCritLayerNum+" ("+botCritDepth.getDepth()+")");
                 }
-                tauBranches[waveNum][critNum] = new TauBranch(topCritDepth.getDepth(),
-                                                              botCritDepth.getDepth(),
-                                                              isPWave);
-                tauBranches[waveNum][critNum].DEBUG = DEBUG;
-                tauBranches[waveNum][critNum].createBranch(sMod,
-                                                           rayParams);
+                tauBranches[waveNum][critNum] =
+                        TauBranch.createBranch(sMod,
+                                rayParams, topCritDepth.getDepth(),
+                                botCritDepth.getDepth(),
+                                isPWave, DEBUG);
                 /*
                  * update minPSoFar. Note that the new minPSoFar could be at the
                  * start of a discontinuty over a high slowness zone, so we need
@@ -460,7 +604,7 @@ public class TauModel implements Serializable {
         double bestCmb = Double.MAX_VALUE;
         double bestIocb = Double.MAX_VALUE;
         for(int branchNum = 0; branchNum < tauBranches[0].length; branchNum++) {
-            TauBranch tBranch = (TauBranch)tauBranches[0][branchNum];
+            TauBranch tBranch = tauBranches[0][branchNum];
             if(Math.abs(tBranch.getTopDepth() - sMod.vMod.getMohoDepth()) <= bestMoho) {
                 mohoBranch = branchNum;
                 bestMoho = Math.abs(tBranch.getTopDepth()
@@ -478,7 +622,7 @@ public class TauModel implements Serializable {
             }
         }
         // check bottom of last layer, zero radius, in case no core for cmb and iocb
-        TauBranch tBranch = (TauBranch)tauBranches[0][tauBranches[0].length-1];
+        TauBranch tBranch = tauBranches[0][tauBranches[0].length-1];
         if(Math.abs(tBranch.getBotDepth() - sMod.vMod.getCmbDepth()) < bestCmb) {
             cmbBranch = tauBranches[0].length;
             cmbDepth = tBranch.getBotDepth();
@@ -561,9 +705,14 @@ public class TauModel implements Serializable {
         if (depthCorrected == null) {
             depthCorrected = splitBranch(depth);
             depthCorrected.sourceDepth = depth;
-            depthCorrected.sourceBranch = depthCorrected.findBranch(depth);
+            /* Check to see if depth is center of earth. */
+            if(depthCorrected.tauBranches[0][depthCorrected.tauBranches[0].length - 1].getBotDepth() == depth) {
+                depthCorrected.sourceBranch = depthCorrected.tauBranches[0].length; // one past bottom
+            } else {
+                depthCorrected.sourceBranch = depthCorrected.findBranch(depth);
+            }
             depthCorrected.validate();
-            depthCache.put(depth, new SoftReference<TauModel>(depthCorrected));
+            depthCache.put(depth, new SoftReference<>(depthCorrected));
         }
         return depthCorrected;
     }
@@ -681,8 +830,8 @@ public class TauModel implements Serializable {
             int branchToSplit = findBranch(depth);
             TauBranch[][] newtauBranches = new TauBranch[2][getNumBranches() + 1];
             for(int i = 0; i < branchToSplit; i++) {
-                newtauBranches[0][i] = (TauBranch)tauBranches[0][i].clone();
-                newtauBranches[1][i] = (TauBranch)tauBranches[1][i].clone();
+                newtauBranches[0][i] = tauBranches[0][i].clone();
+                newtauBranches[1][i] = tauBranches[1][i].clone();
                 if(indexS != -1) {
                     // add the new ray parameter from splitting the S Wave
                     // slowness layer to both the P and S wave Tau branches
@@ -697,11 +846,12 @@ public class TauModel implements Serializable {
                 }
             }
             for(int pOrS = 0; pOrS < 2; pOrS++) {
-                newtauBranches[pOrS][branchToSplit] = new TauBranch(tauBranches[pOrS][branchToSplit].getTopDepth(),
-                                                                    depth,
-                                                                    pOrS == 0);
-                newtauBranches[pOrS][branchToSplit].createBranch(outSMod,
-                                                                 outRayParams);
+                newtauBranches[pOrS][branchToSplit] =
+                        TauBranch.createBranch(outSMod,
+                                outRayParams,
+                                tauBranches[pOrS][branchToSplit].getTopDepth(),
+                                depth,
+                                pOrS == 0, DEBUG);
                 newtauBranches[pOrS][branchToSplit + 1] = tauBranches[pOrS][branchToSplit].difference(newtauBranches[pOrS][branchToSplit],
                                                                                                       indexP,
                                                                                                       indexS,
@@ -710,7 +860,7 @@ public class TauModel implements Serializable {
             }
             for(int i = branchToSplit + 1; i < tauBranches[0].length; i++) {
                 for(int pOrS = 0; pOrS < 2; pOrS++) {
-                    newtauBranches[pOrS][i + 1] = (TauBranch)tauBranches[pOrS][i].clone();
+                    newtauBranches[pOrS][i + 1] = tauBranches[pOrS][i].clone();
                 }
                 if(indexS != -1) {
                     // add the new ray parameter from splitting the S Wave
@@ -798,9 +948,8 @@ public class TauModel implements Serializable {
     }
 
     public static TauModel readModel(String filename)
-            throws FileNotFoundException, IOException,
-            StreamCorruptedException, ClassNotFoundException,
-            OptionalDataException {
+            throws IOException,
+            ClassNotFoundException {
         TauModel tMod;
         BufferedInputStream in = new BufferedInputStream(new FileInputStream(filename));
         try {
@@ -812,12 +961,10 @@ public class TauModel implements Serializable {
     }
 
     public static TauModel readModelFromStream(InputStream inStream)
-            throws InvalidClassException, IOException,
-            StreamCorruptedException, ClassNotFoundException,
-            OptionalDataException {
+            throws IOException,
+            ClassNotFoundException {
         ObjectInputStream in = new ObjectInputStream(inStream);
-        TauModel tMod = (TauModel)in.readObject();
-        return tMod;
+        return (TauModel)in.readObject();
     }
 
     /*
@@ -863,7 +1010,7 @@ public class TauModel implements Serializable {
     public boolean validate() {
         for(int i = 0; i < rayParams.length - 1; i++) {
             if(rayParams[i + 1] >= rayParams[i]) {
-                System.err.println("RayParams are not monotonically decreasing. "
+                Alert.warning("RayParams are not monotonically decreasing. "
                         + "rayParams["
                         + i
                         + "]="
@@ -873,36 +1020,36 @@ public class TauModel implements Serializable {
             }
         }
         if(tauBranches[0].length != tauBranches[1].length) {
-            System.err.println("TauBranches for P and S are not equal. "
+            Alert.warning("TauBranches for P and S are not equal. "
                     + tauBranches[0].length + " " + tauBranches[1].length);
             return false;
         }
         if(tauBranches[0][0].getTopDepth() != 0
                 || tauBranches[1][0].getTopDepth() != 0) {
-            System.err.println("branch 0 topDepth != 0");
+            Alert.warning("branch 0 topDepth != 0");
             return false;
         }
         for(int i = 1; i < getNumBranches(); i++) {
             if(tauBranches[0][i].getTopDepth() != tauBranches[1][i].getTopDepth()) {
-                System.err.println("branch " + i + " P topDepth != S topDepth");
+                Alert.warning("branch " + i + " P topDepth != S topDepth");
                 return false;
             }
             if(tauBranches[0][i].getBotDepth() != tauBranches[1][i].getBotDepth()) {
-                System.err.println("branch " + i + " P botDepth != S botDepth");
+                Alert.warning("branch " + i + " P botDepth != S botDepth");
                 return false;
             }
             if(tauBranches[0][i].getTopDepth() != tauBranches[0][i - 1].getBotDepth()) {
-                System.err.println("branch " + i + " topDepth != botDepth of "
+                Alert.warning("branch " + i + " topDepth != botDepth of "
                         + (i - 1));
                 return false;
             }
         }
         if(tauBranches[0][getNumBranches() - 1].getMinRayParam() != 0) {
-            System.err.println("branch tauBranches[0].length-1 minRayParam != 0");
+            Alert.warning("branch tauBranches[0].length-1 minRayParam != 0");
             return false;
         }
         if(tauBranches[1][getNumBranches() - 1].getMinRayParam() != 0) {
-            System.err.println("branch tauBranches[1].length-1 minRayParam != 0");
+            Alert.warning("branch tauBranches[1].length-1 minRayParam != 0");
             return false;
         }
         return true;
@@ -911,28 +1058,45 @@ public class TauModel implements Serializable {
     public void print() {
         double deg, time;
         if(DEBUG)
-            System.out.println("Starting print() in TauModel");
-        System.out.println("Delta tau for each slowness sample and layer.");
+            Alert.debug("Starting print() in TauModel");
+        Alert.debug("Delta tau for each slowness sample and layer.");
         for(int j = 0; j < rayParams.length; j++) {
             deg = 0;
             time = 0;
             for(int i = 0; i < getNumBranches(); i++) {
                 deg += tauBranches[0][i].getDist(j) * 180 / Math.PI;
                 time += tauBranches[0][i].time[j];
-                System.out.println(" i " + i + " j " + j + " rayParam "
+                Alert.debug(" i " + i + " j " + j + " rayParam "
                         + rayParams[j] + " tau " + tauBranches[0][i].tau[j]
                         + " time " + tauBranches[0][i].time[j] + " dist "
                         + tauBranches[0][i].getDist(j) + " degrees "
                         + (tauBranches[0][i].getDist(j) * 180 / Math.PI));
             }
-            System.out.println();
-            System.out.println("deg= " + deg + "  time=" + time);
+            Alert.debug("");
+            Alert.debug("deg= " + deg + "  time=" + time);
         }
     }
 
     public String toString() {
+        return branchesToString().toString();
+    }
+
+    public StringBuilder branchesToString() {
+        StringBuilder sb = new StringBuilder(getModelName()).append("\n");
+        for(int i = 0; i < tauBranches[0].length; i++) {
+            TauBranch pBranch = tauBranches[0][i];
+            TauBranch sBranch = tauBranches[1][i];
+            sb.append(i).append(" P ").append(pBranch.getTopDepth()).append(" ").append(pBranch.getBotDepth());
+            sb.append(" ").append(pBranch.getTopRayParam()).append(" ").append(pBranch.getBotRayParam()).append("\n");
+            sb.append(i).append(" S ").append(sBranch.getTopDepth()).append(" ").append(sBranch.getBotDepth());
+            sb.append(" ").append(sBranch.getTopRayParam()).append(" ").append(sBranch.getBotRayParam()).append("\n");
+        }
+        return sb;
+    }
+
+    public String dumpToString() {
         if(DEBUG)
-            System.out.println("Starting toString() in TauModel");
+            Alert.debug("Starting toString() in TauModel");
         String desc = "Delta tau for each slowness sample and layer.\n";
         for(int j = 0; j < rayParams.length; j++) {
             for(int i = 0; i < tauBranches[0].length; i++) {
@@ -959,6 +1123,6 @@ public class TauModel implements Serializable {
         return null;
     }
     
-    private HashMap<Double, SoftReference<TauModel>> depthCache = new HashMap<Double, SoftReference<TauModel>>();
+    private final HashMap<Double, SoftReference<TauModel>> depthCache = new HashMap<>();
 
 }
