@@ -29,25 +29,10 @@
  */
 package edu.sc.seis.TauP;
 
-import org.json.JSONArray;
-import org.json.JSONObject;
-
-import java.io.BufferedWriter;
-import java.io.File;
-import java.io.FileReader;
-import java.io.FileWriter;
-import java.io.IOException;
-import java.io.OutputStreamWriter;
-import java.io.PrintWriter;
-import java.io.Reader;
-import java.io.Serializable;
-import java.io.StreamTokenizer;
-import java.io.Writer;
+import java.io.*;
 import java.util.ArrayList;
 import java.util.List;
-
-import static edu.sc.seis.TauP.PhaseInteraction.TRANSDOWN;
-import static edu.sc.seis.TauP.PhaseInteraction.TRANSUP;
+import java.util.Objects;
 
 /**
  * This class defines basic classes to store and manipulate a velocity model.
@@ -60,6 +45,9 @@ import static edu.sc.seis.TauP.PhaseInteraction.TRANSUP;
  */
 public class VelocityModel implements Cloneable, Serializable {
 
+    public static final String ND = "nd";
+    public static final String TVEL = "tvel";
+
     public VelocityModel(String modelName,
                          double radiusOfEarth,
                          double mohoDepth,
@@ -155,7 +143,7 @@ public class VelocityModel implements Cloneable, Serializable {
 
     /** set the model name. */
     public void setModelName(String modelName) {
-        if(modelName.length() > 0) {
+        if(!modelName.isEmpty()) {
             this.modelName = modelName;
         } else {
             this.modelName = "unknown";
@@ -185,6 +173,28 @@ public class VelocityModel implements Cloneable, Serializable {
         }
         return false;
     }
+
+    public List<NamedVelocityDiscon> getNamedDiscons() {
+        return namedDiscon;
+    }
+
+    public NamedVelocityDiscon getNamedDisconForDepth(double depth) {
+        for (NamedVelocityDiscon nDiscon : namedDiscon) {
+            if (depth == nDiscon.depth) {
+                return nDiscon;
+            }
+        }
+        return null;
+    }
+
+    public boolean isNamedDisconDepth(double depth) {
+        for (NamedVelocityDiscon nDiscon : namedDiscon) {
+            if (depth == nDiscon.depth) {
+                return true;
+            }
+        }
+        return false;
+    }
     
     /** @return the depths of discontinuities within the velocity model */
     public double[] getDisconDepths() {
@@ -279,6 +289,22 @@ public class VelocityModel implements Cloneable, Serializable {
         this.spherical = spherical;
     }
 
+    public boolean densityIsDefault() {
+        boolean allDensityDefault = true;
+        for(VelocityLayer vlay: getLayers() ) {
+            allDensityDefault &= vlay.densityIsDefault();
+        }
+        return allDensityDefault;
+    }
+
+    public boolean QIsDefault() {
+        boolean allQDefault = true;
+        for(VelocityLayer vlay: getLayers() ) {
+            allQDefault &= vlay.QIsDefault();
+        }
+        return allQDefault;
+    }
+
     public VelocityLayer getVelocityLayerClone(int layerNum) {
         return (VelocityLayer)(layer.get(layerNum)).clone();
     }
@@ -293,7 +319,7 @@ public class VelocityModel implements Cloneable, Serializable {
     }
     
     public VelocityLayer[] getLayers() {
-        return (VelocityLayer[])layer.toArray(new VelocityLayer[0]);
+        return layer.toArray(new VelocityLayer[0]);
     }
 
     // normal methods
@@ -308,7 +334,7 @@ public class VelocityModel implements Cloneable, Serializable {
     public int layerNumberAbove(double depth) throws NoSuchLayerException {
         VelocityLayer tempLayer;
         /* first check to see if depth is at top of top layer. */
-        tempLayer = (VelocityLayer)getVelocityLayer(0);
+        tempLayer = getVelocityLayer(0);
         if(depth == tempLayer.getTopDepth()) {
             return 0;
         } else {
@@ -383,13 +409,11 @@ public class VelocityModel implements Cloneable, Serializable {
      * @return the value of the given material property
      * @exception NoSuchLayerException
      *                occurs if no layer contains the given depth.
-     * @exception NoSuchMatPropException
-     *                occurs if the material property is not recognized.
      */
-    public double evaluateAbove(double depth, char materialProperty)
-            throws NoSuchLayerException, NoSuchMatPropException {
+    public double evaluateAbove(double depth, VelocityModelMaterial materialProperty)
+            throws NoSuchLayerException {
         VelocityLayer tempLayer;
-        tempLayer = (VelocityLayer)getVelocityLayer(layerNumberAbove(depth));
+        tempLayer = getVelocityLayer(layerNumberAbove(depth));
         return tempLayer.evaluateAt(depth, materialProperty);
     }
 
@@ -401,13 +425,11 @@ public class VelocityModel implements Cloneable, Serializable {
      * @return the value of the given material property
      * @exception NoSuchLayerException
      *                occurs if no layer contains the given depth.
-     * @exception NoSuchMatPropException
-     *                occurs if the material property is not recognized.
      */
-    public double evaluateBelow(double depth, char materialProperty)
-            throws NoSuchLayerException, NoSuchMatPropException {
+    public double evaluateBelow(double depth, VelocityModelMaterial materialProperty)
+            throws NoSuchLayerException {
         VelocityLayer tempLayer;
-        tempLayer = (VelocityLayer)getVelocityLayer(layerNumberBelow(depth));
+        tempLayer = getVelocityLayer(layerNumberBelow(depth));
         return tempLayer.evaluateAt(depth, materialProperty);
     }
 
@@ -416,13 +438,11 @@ public class VelocityModel implements Cloneable, Serializable {
      * velocity, at the top of the given layer.
      * 
      * @return the value of the given material property
-     * @exception NoSuchMatPropException
      *                occurs if the material property is not recognized.
      */
-    public double evaluateAtTop(int layerNumber, char materialProperty)
-            throws NoSuchMatPropException {
+    public double evaluateAtTop(int layerNumber, VelocityModelMaterial materialProperty) {
         VelocityLayer tempLayer;
-        tempLayer = (VelocityLayer)getVelocityLayer(layerNumber);
+        tempLayer = getVelocityLayer(layerNumber);
         return tempLayer.evaluateAtTop(materialProperty);
     }
 
@@ -431,13 +451,10 @@ public class VelocityModel implements Cloneable, Serializable {
      * velocity, at the bottom of the given layer.
      * 
      * @return the value of the given material property
-     * @exception NoSuchMatPropException
-     *                occurs if the material property is not recognized.
      */
-    public double evaluateAtBottom(int layerNumber, char materialProperty)
-            throws NoSuchMatPropException {
+    public double evaluateAtBottom(int layerNumber, VelocityModelMaterial materialProperty) {
         VelocityLayer tempLayer;
-        tempLayer = (VelocityLayer)getVelocityLayer(layerNumber);
+        tempLayer = getVelocityLayer(layerNumber);
         return tempLayer.evaluateAtBottom(materialProperty);
     }
 
@@ -448,7 +465,7 @@ public class VelocityModel implements Cloneable, Serializable {
      */
     public double depthAtTop(int layerNumber) {
         VelocityLayer tempLayer;
-        tempLayer = (VelocityLayer)getVelocityLayer(layerNumber);
+        tempLayer = getVelocityLayer(layerNumber);
         return tempLayer.getTopDepth();
     }
 
@@ -456,12 +473,11 @@ public class VelocityModel implements Cloneable, Serializable {
      * returns the depth at the bottom of the given layer.
      * 
      * @return the depth.
-     * @exception NoSuchMatPropException
      *                occurs if the material property is not recognized.
      */
-    public double depthAtBottom(int layerNumber) throws NoSuchMatPropException {
+    public double depthAtBottom(int layerNumber) {
         VelocityLayer tempLayer;
-        tempLayer = (VelocityLayer)getVelocityLayer(layerNumber);
+        tempLayer = getVelocityLayer(layerNumber);
         return tempLayer.getBotDepth();
     }
 
@@ -477,8 +493,8 @@ public class VelocityModel implements Cloneable, Serializable {
                                        boolean smoothTop,
                                        boolean smoothBot)
             throws VelocityModelException {
-        try {
-            List<VelocityLayer> outLayers = new ArrayList<VelocityLayer>();
+
+            List<VelocityLayer> outLayers = new ArrayList<>();
             int topLayerNum = layerNumberBelow(newLayers[0].getTopDepth());
             int numAdded = 0;
             for(int i = 0; i<topLayerNum; i++) {
@@ -495,20 +511,29 @@ public class VelocityModel implements Cloneable, Serializable {
                                                 topLayer.getTopDepth(),
                                                 newLayers[0].getTopDepth(),
                                                 topLayer.getTopPVelocity(),
-                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), 'P'),
+                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.P_VELOCITY),
                                                 topLayer.getTopSVelocity(),
-                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), 'S'),
+                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.S_VELOCITY),
                                                 topLayer.getTopDensity(),
-                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), 'R')));
+                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.DENSITY),
+                        topLayer.getTopQp(),
+                        topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.Q_P),
+                        topLayer.getTopQs(),
+                        topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.Q_S)
+                        ));
                 outLayers.add(new VelocityLayer(numAdded++,
                                                 newLayers[0].getTopDepth(),
                                                 topLayer.getBotDepth(),
-                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), 'P'),
+                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.P_VELOCITY),
                                                 topLayer.getBotPVelocity(),
-                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), 'S'),
+                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.S_VELOCITY),
                                                 topLayer.getBotSVelocity(),
-                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), 'R'),
-                                                topLayer.getBotDensity()));
+                                                topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.DENSITY),
+                                                topLayer.getBotDensity(),
+                        topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.Q_P),
+                        topLayer.getBotQp(),
+                        topLayer.evaluateAt(newLayers[0].getTopDepth(), VelocityModelMaterial.Q_S),
+                        topLayer.getBotQs()));
             } else {
                 // already a discon at our new layer top depth
                 outLayers.add(topLayer.cloneRenumber(numAdded++));
@@ -524,19 +549,19 @@ public class VelocityModel implements Cloneable, Serializable {
                                                 botLayer.getTopDepth(),
                                                 lastNewLayer.getBotDepth(),
                                                 botLayer.getTopPVelocity(),
-                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), 'P'),
+                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), VelocityModelMaterial.P_VELOCITY),
                                                 botLayer.getTopSVelocity(),
-                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), 'S'),
+                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), VelocityModelMaterial.S_VELOCITY),
                                                 botLayer.getTopDensity(),
-                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), 'R')));
+                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), VelocityModelMaterial.DENSITY)));
                 outLayers.add(new VelocityLayer(numAdded++,
                                                 lastNewLayer.getBotDepth(),
                                                 botLayer.getBotDepth(),
-                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), 'P'),
+                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), VelocityModelMaterial.P_VELOCITY),
                                                 botLayer.getBotPVelocity(),
-                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), 'S'),
+                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), VelocityModelMaterial.S_VELOCITY),
                                                 botLayer.getBotSVelocity(),
-                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), 'R'),
+                                                botLayer.evaluateAt(lastNewLayer.getBotDepth(), VelocityModelMaterial.DENSITY),
                                                 botLayer.getBotDensity()));
             } else {
                 // already a discon at our new layer top depth
@@ -568,7 +593,7 @@ public class VelocityModel implements Cloneable, Serializable {
                 }
             }
             
-            List<VelocityLayer> replaceoutLayers = new ArrayList<VelocityLayer>();
+            List<VelocityLayer> replaceoutLayers = new ArrayList<>();
             numAdded = 0;
             for(VelocityLayer vlay : outLayers) {
                 if (vlay.getTopDepth() < newLayers[0].getTopDepth()) {
@@ -611,10 +636,6 @@ public class VelocityModel implements Cloneable, Serializable {
                 throw new VelocityModelException("replace layers but now is not valid.");
             }
             return outVMod;
-        } catch(NoSuchMatPropException e) {
-            // can't happen, but...
-            throw new RuntimeException(e);
-        }
     }
 
 
@@ -625,8 +646,7 @@ public class VelocityModel implements Cloneable, Serializable {
     public VelocityModel elevationLayer(float elevation,
                                        String name)
             throws VelocityModelException {
-        try {
-            List<VelocityLayer> outLayers = new ArrayList<VelocityLayer>();
+            List<VelocityLayer> outLayers = new ArrayList<>();
 
             int numAdded = 0;
             VelocityLayer elevationLayer = getVelocityLayer(0).cloneRenumber(numAdded);
@@ -656,10 +676,6 @@ public class VelocityModel implements Cloneable, Serializable {
                 throw new VelocityModelException("replace layers but now is not valid.");
             }
             return outVMod;
-        } catch(NoSuchMatPropException e) {
-            // can't happen, but...
-            throw new RuntimeException(e);
-        }
     }
 
     /**
@@ -668,7 +684,7 @@ public class VelocityModel implements Cloneable, Serializable {
      */
     public void printGMT(String filename) throws IOException {
         String psFile;
-        if (filename == "stdout") { 
+        if (Objects.equals(filename, "stdout")) {
             psFile = "taup_velocitymodel";
         } else if(filename.endsWith(".gmt")) {
             psFile = filename.substring(0, filename.length() - 4) + ".ps";
@@ -677,7 +693,7 @@ public class VelocityModel implements Cloneable, Serializable {
         }
 
         PrintWriter dos;
-        if (filename == "stdout") {
+        if (filename.equals("stdout")) {
             dos = new PrintWriter(new OutputStreamWriter(System.out));
         } else {
             dos = new PrintWriter(new BufferedWriter(new FileWriter(filename)));
@@ -722,14 +738,14 @@ public class VelocityModel implements Cloneable, Serializable {
      * prints out the velocity model into a file in a for suitable for plotting
      * with GMT.
      */
-    public void printGMT(PrintWriter dos) throws IOException {
+    public void printGMT(PrintWriter dos) {
         dos.println("> P velocity for " + modelName + "  below");
         printGMTforP(dos);
         dos.println("> S velocity for " + modelName + "  below");
         printGMTforP(dos);
     }
     
-    void printGMTforP(PrintWriter dos) throws IOException {
+    void printGMTforP(PrintWriter dos) {
         double pVel = -1.0;
         for(int layerNum = 0; layerNum < getNumLayers(); layerNum++) {
             VelocityLayer currVelocityLayer = getVelocityLayer(layerNum);
@@ -743,7 +759,7 @@ public class VelocityModel implements Cloneable, Serializable {
         }
     }
     
-    void printGMTforS(PrintWriter dos) throws IOException {
+    void printGMTforS(PrintWriter dos) {
         double sVel = -1.0;
         for(int layerNum = 0; layerNum < getNumLayers(); layerNum++) {
             VelocityLayer currVelocityLayer = getVelocityLayer(layerNum);
@@ -764,55 +780,55 @@ public class VelocityModel implements Cloneable, Serializable {
         VelocityLayer currVelocityLayer, prevVelocityLayer;
         /* is radiusOfEarth positive? */
         if(radiusOfEarth <= 0.0) {
-            System.err.println("Radius of earth is not positive. radiusOfEarth = "
+            Alert.warning("Radius of earth is not positive. radiusOfEarth = "
                     + radiusOfEarth);
             return false;
         }
         /* is mohoDepth non-negative? */
         if(mohoDepth < 0.0) {
-            System.err.println("mohoDepth is not non-negative. mohoDepth = "
+            Alert.warning("mohoDepth is not non-negative. mohoDepth = "
                     + mohoDepth);
             return false;
         }
         /* is cmbDepth >= mohoDepth? */
         if(cmbDepth < mohoDepth) {
-            System.err.println("cmbDepth < mohoDepth. cmbDepth = " + cmbDepth
+            Alert.warning("cmbDepth < mohoDepth. cmbDepth = " + cmbDepth
                     + " mohoDepth = " + mohoDepth);
             return false;
         }
         /* is cmbDepth non-negative? */
         if(cmbDepth < 0.0) {
-            System.err.println("cmbDepth is negative. cmbDepth = "
+            Alert.warning("cmbDepth is negative. cmbDepth = "
                     + cmbDepth);
             return false;
         }
         /* is iocbDepth >= cmbDepth? */
         if(iocbDepth < cmbDepth) {
-            System.err.println("iocbDepth < cmbDepth. iocbDepth = " + iocbDepth
+            Alert.warning("iocbDepth < cmbDepth. iocbDepth = " + iocbDepth
                     + " cmbDepth = " + cmbDepth);
             return false;
         }
         /* is iocbDepth non-negative? */
         if(iocbDepth < 0.0) {
-            System.err.println("iocbDepth is negative. iocbDepth = "
+            Alert.warning("iocbDepth is negative. iocbDepth = "
                     + iocbDepth);
             return false;
         }
         /* is minRadius non-negative? */
         if(minRadius < 0.0) {
-            System.err.println("minRadius is not non-negative. minRadius = "
+            Alert.warning("minRadius is not non-negative. minRadius = "
                     + minRadius);
             return false;
         }
         /* is maxRadius positive? */
         if(maxRadius <= 0.0) {
-            System.err.println("maxRadius is not positive. maxRadius = "
+            Alert.warning("maxRadius is not positive. maxRadius = "
                     + maxRadius);
             return false;
         }
         /* is maxRadius > minRadius? */
         if(maxRadius <= minRadius) {
-            System.err.println("maxRadius <= minRadius. maxRadius = "
+            Alert.warning("maxRadius <= minRadius. maxRadius = "
                     + maxRadius + " minRadius = " + minRadius);
             return false;
         }
@@ -832,21 +848,21 @@ public class VelocityModel implements Cloneable, Serializable {
                 /*
                  * There is a gap in the velocity model!
                  */
-                System.err.println("There is a gap in the velocity model "
+                Alert.warning("There is a gap in the velocity model "
                         + "between layers " + (layerNum - 1) + " and "
                         + layerNum);
-                System.err.println("prevVelocityLayer=" + prevVelocityLayer);
-                System.err.println("currVelocityLayer=" + currVelocityLayer);
+                Alert.warning("prevVelocityLayer=" + prevVelocityLayer);
+                Alert.warning("currVelocityLayer=" + currVelocityLayer);
                 return false;
             }
             if(currVelocityLayer.getBotDepth() == currVelocityLayer.getTopDepth()) {
                 /*
                  * This layer has zero thickness.
                  */
-                System.err.println("There is a zero thickness layer in the "
+                Alert.warning("There is a zero thickness layer in the "
                         + "velocity model at layer " + layerNum);
-                System.err.println("prevVelocityLayer=" + prevVelocityLayer);
-                System.err.println("currVelocityLayer=" + currVelocityLayer);
+                Alert.warning("prevVelocityLayer=" + prevVelocityLayer);
+                Alert.warning("currVelocityLayer=" + currVelocityLayer);
                 return false;
             }
             if(currVelocityLayer.getTopPVelocity() <= 0.0
@@ -854,7 +870,7 @@ public class VelocityModel implements Cloneable, Serializable {
                 /*
                  * This layer has a negative or zero P velocity.
                  */
-                System.err.println("There is a negative P velocity layer in the "
+                Alert.warning("There is a negative P velocity layer in the "
                         + "velocity model at layer " + layerNum);
                 return false;
             }
@@ -863,7 +879,7 @@ public class VelocityModel implements Cloneable, Serializable {
                 /*
                  * This layer has a negative S velocity.
                  */
-                System.err.println("There is a negative S velocity layer in the "
+                Alert.warning("There is a negative S velocity layer in the "
                         + "velocity model at layer " + layerNum);
                 return false;
             }
@@ -872,7 +888,7 @@ public class VelocityModel implements Cloneable, Serializable {
                 /*
                  * This layer goes to zero P velocity without a discontinuity.
                  */
-                System.err.println("There is a layer that goes to zero P velocity "
+                Alert.warning("There is a layer that goes to zero P velocity "
                         + "without a discontinuity in the "
                         + "velocity model at layer "
                         + layerNum
@@ -886,7 +902,7 @@ public class VelocityModel implements Cloneable, Serializable {
                 /*
                  * This layer goes to zero S velocity without a discontinuity.
                  */
-                System.err.println("There is a layer that goes to zero S velocity "
+                Alert.warning("There is a layer that goes to zero S velocity "
                         + "without a discontinuity in the "
                         + "velocity model at layer "
                         + layerNum
@@ -912,22 +928,35 @@ public class VelocityModel implements Cloneable, Serializable {
 
     public void writeToND(Writer out) throws IOException {
         VelocityLayer prev = null;
+        boolean allDensityDefault = densityIsDefault();
+        boolean allQDefault = QIsDefault();
         for(VelocityLayer vlay: getLayers() ) {
-            if (prev == null || 
+            if (prev == null ||
+                    vlay.getBotDepth() == getMohoDepth() ||
+                    vlay.getBotDepth() == getCmbDepth() ||
+                    vlay.getBotDepth() == getIocbDepth() ||
                     prev.getBotPVelocity()!=vlay.getTopPVelocity() ||
                     prev.getBotSVelocity()!=vlay.getTopSVelocity() ||
                     prev.getBotDensity()!=vlay.getTopDensity()) {
-                out.write(vlay.getTopDepth()+" "+vlay.getTopPVelocity()+" "+vlay.getTopSVelocity()+" "+vlay.getTopDensity()+"\n");
+                out.write(vlay.getTopDepth()+" "+vlay.getTopPVelocity()+" "+vlay.getTopSVelocity());
+                if (!allDensityDefault) {
+                    out.write(" "+vlay.getTopDensity());
+                }
+                if (!allQDefault) {
+                    out.write(" "+vlay.getTopQp()+" "+vlay.getTopQs());
+                }
+                out.write("\n");
             }
-            out.write(vlay.getBotDepth()+" "+vlay.getBotPVelocity()+" "+vlay.getBotSVelocity()+" "+vlay.getBotDensity()+"\n");
-            if (vlay.getBotDepth() == getMohoDepth()) {
-                out.write("mantle\n");
+            out.write(vlay.getBotDepth()+" "+vlay.getBotPVelocity()+" "+vlay.getBotSVelocity());
+            if (!allDensityDefault) {
+                out.write(" "+vlay.getBotDensity());
             }
-            if (vlay.getBotDepth() == getCmbDepth()) {
-                out.write("outer-core\n");
+            if (!allQDefault) {
+                out.write(" "+vlay.getBotQp()+" "+vlay.getBotQs());
             }
-            if (vlay.getBotDepth() == getIocbDepth()) {
-                out.write("inner-core\n");
+            out.write("\n");
+            if (isNamedDisconDepth(vlay.getBotDepth()) ) {
+                out.write(getNamedDisconForDepth(vlay.getBotDepth()).getName()+"\n");
             }
             prev = vlay;
         }
@@ -939,66 +968,6 @@ public class VelocityModel implements Cloneable, Serializable {
         }
     }
 
-    public JSONObject asJSON() {
-        JSONObject json = new JSONObject();
-        json.put("modelName", getModelName());
-        json.put("radiusOfEarth", getRadiusOfEarth());
-        json.put("mohoDepth", getMohoDepth());
-        json.put("cmbDepth", getCmbDepth());
-        json.put("iocbDepth", getIocbDepth());
-        json.put("minRadius", getMinRadius());
-        json.put("maxRadius", getMaxRadius());
-        json.put("spherical", getSpherical());
-        JSONArray layers = new JSONArray();
-        json.put("layers", layers);
-        for (VelocityLayer vl : getLayers()) {
-            layers.put(vl.asJSON());
-        }
-        return json;
-    }
-
-    public String asJSON(boolean pretty, String indent) {
-        String NL = "";
-        if (pretty) {
-            NL = "\n";
-        }
-        String Q = ""+'"';
-        String COMMA = ",";
-        String QCOMMA = Q+COMMA;
-        String COLON = ": "; // plus space
-        String S = "  ";
-        String QC = Q+COLON;
-        String QCQ = QC+Q;
-        String SS = S+S;
-        String SQ = S+Q;
-        String SSQ = S+SQ;
-        StringBuilder out = new StringBuilder();
-        out.append(indent+"{"+NL);
-        out.append(indent+SQ+"modelName"+QC+Q+modelName+Q+COMMA+NL);
-        out.append(indent+SQ+"radiusOfEarth"+QC+radiusOfEarth+COMMA+NL);
-        out.append(indent+SQ+"mohoDepth"+QC+mohoDepth+COMMA+NL);
-        out.append(indent+SQ+"cmbDepth"+QC+cmbDepth+COMMA+NL);
-        out.append(indent+SQ+"iocbDepth"+QC+iocbDepth+COMMA+NL);
-        out.append(indent+SQ+"minRadius"+QC+minRadius+COMMA+NL);
-        out.append(indent+SQ+"maxRadius"+QC+maxRadius+COMMA+NL);
-        out.append(indent+SQ+"spherical"+QC+spherical+COMMA+NL);
-        out.append(indent+SQ+"numLayers"+QC+getNumLayers()+COMMA+NL);
-        out.append(indent+SQ+"layers"+QC+"[");
-        boolean first = true;
-        for (VelocityLayer vl : getLayers()) {
-            if (first) {
-                first = false;
-                out.append(NL);
-            } else {
-                out.append(COMMA+NL);
-            }
-            out.append(vl.asJSON( pretty, indent));
-        }
-        out.append(indent+S+"]"+NL);
-        out.append(indent+"}");
-        return out.toString();
-    }
-
     public static String getModelNameFromFileName(String filename) {
         int j = filename.lastIndexOf(System.getProperty("file.separator"));
         String modelFilename = filename.substring(j + 1);
@@ -1008,7 +977,7 @@ public class VelocityModel implements Cloneable, Serializable {
         } else if(modelFilename.endsWith(".nd")) {
             modelName = modelFilename.substring(0, modelFilename.length() - 3);
         } else if(modelFilename.startsWith("GB.")) {
-            modelName = modelFilename.substring(3, modelFilename.length());
+            modelName = modelFilename.substring(3);
         } else {
             modelName = modelFilename;
         }
@@ -1025,7 +994,7 @@ public class VelocityModel implements Cloneable, Serializable {
     public static VelocityModel readVelocityFile(String filename,
                                                  String fileType)
             throws IOException, VelocityModelException {
-        if (fileType == null || fileType.equals("")) {
+        if (fileType == null || fileType.isEmpty()) {
             if (filename.endsWith(".nd")) {
                 fileType = ".nd";
             } else if (filename.endsWith(".tvel")) {
@@ -1033,7 +1002,7 @@ public class VelocityModel implements Cloneable, Serializable {
             }
         }
         if (fileType.startsWith(".")) {
-            fileType = fileType.substring(1, fileType.length());
+            fileType = fileType.substring(1);
         }
         File f = new File(filename);
         if ( ! f.exists() && ! filename.endsWith("."+fileType) && new File(filename+"."+fileType).exists()) {
@@ -1133,7 +1102,7 @@ public class VelocityModel implements Cloneable, Serializable {
             tokenIn.nextToken();
         }
         VelocityLayer lastZeroThickLayer = null;
-        List<VelocityLayer> layers = new ArrayList<VelocityLayer>();
+        List<VelocityLayer> layers = new ArrayList<>();
         while(tokenIn.ttype != StreamTokenizer.TT_EOF) {
             // Loop until we hit the end of file
             botDepth = tokenIn.nval;
@@ -1209,6 +1178,11 @@ public class VelocityModel implements Cloneable, Serializable {
                                  true,
                                  layers);
         vMod.fixDisconDepths();
+        if (vMod.namedDiscon.isEmpty()) {
+            vMod.namedDiscon.add(new NamedVelocityDiscon(NamedVelocityDiscon.MOHO, vMod.getMohoDepth()));
+            vMod.namedDiscon.add(new NamedVelocityDiscon(NamedVelocityDiscon.CMB, vMod.getCmbDepth()));
+            vMod.namedDiscon.add(new NamedVelocityDiscon(NamedVelocityDiscon.IOCB, vMod.getIocbDepth()));
+        }
         return vMod;
     }
 
@@ -1222,15 +1196,15 @@ public class VelocityModel implements Cloneable, Serializable {
      * This feature makes phase interpretation much easier to
      * code. Also, as they are not needed for travel time calculations, the
      * density, Qp and Qs may be omitted.
-     * 
+     * <p>
      * The velocities are assumed to be linear between sample points. Because
      * this type of model file doesn't give complete information we make the
      * following assumptions: 
-     * 
+     * <p>
      * modelname - from the filename, with ".nd" dropped, if present 
-     * 
+     * <p>
      * radiusOfEarth - the largest depth in the model
-     * 
+     * <p>
      * Also, because this method makes use of the string tokenizer, comments are
      * allowed. # as well as // signify that the rest of the line is a comment.
      * C style slash-star comments are also allowed.
@@ -1282,7 +1256,7 @@ public class VelocityModel implements Cloneable, Serializable {
         int myLayerNumber = 0;
         VelocityLayer tempLayer;
         VelocityLayer lastZeroThickLayer = null;
-        double topDepth, topPVel, topSVel, topDensity = 2.6, topQp = 1000, topQs = 2000;
+        double topDepth, topPVel, topSVel, topDensity = 2.6, topQp = VelocityLayer.DEFAULT_QP, topQs = VelocityLayer.DEFAULT_QS;
         double botDepth, botPVel, botSVel, botDensity = topDensity, botQp = topQp, botQs = topQs;
         List<NamedVelocityDiscon> namedDiscon = new ArrayList<>();
         double mohoDepth = DEFAULT_MOHO;
@@ -1333,7 +1307,7 @@ public class VelocityModel implements Cloneable, Serializable {
         } else {
             tokenIn.nextToken();
         }
-        List<VelocityLayer> layers = new ArrayList<VelocityLayer>();
+        List<VelocityLayer> layers = new ArrayList<>();
         while(tokenIn.ttype != StreamTokenizer.TT_EOF) {
             // Loop until we hit the end of file
             if (tokenIn.ttype == StreamTokenizer.TT_EOL) {
@@ -1519,7 +1493,7 @@ public class VelocityModel implements Cloneable, Serializable {
             iocbMin = Math.abs(iocbDepth - belowLayer.getBotDepth());
         }
         // make sure all named discontinuties are actual discontinuities
-        if (unverifiedNamedDiscon.size() != 0) {
+        if (!unverifiedNamedDiscon.isEmpty()) {
             throw new VelocityModelException("Named Discontinuity not at actual model discontinuity: "+unverifiedNamedDiscon.get(0));
         }
         for (NamedVelocityDiscon nl : namedDiscon) {
@@ -1549,13 +1523,11 @@ public class VelocityModel implements Cloneable, Serializable {
      * model via the earth flattening transform.
      * 
      * @return the flattened VelocityModel object.
-     * @exception VelocityModelException
-     *                occurs ???.
      */
-    public VelocityModel earthFlattenTransform() throws VelocityModelException {
+    public VelocityModel earthFlattenTransform() {
         VelocityLayer newLayer, oldLayer;
         boolean spherical = false;
-        List<VelocityLayer> layers = new ArrayList<VelocityLayer>(vectorLength);
+        List<VelocityLayer> layers = new ArrayList<>(vectorLength);
         for(int i = 0; i < getNumLayers(); i++) {
             oldLayer = getVelocityLayer(i);
             newLayer = new VelocityLayer(i,
@@ -1589,21 +1561,30 @@ public class VelocityModel implements Cloneable, Serializable {
                                  spherical,
                                  layers);
     }
-    
+
+    public ReflTrans calcReflTransCoefFreeSurface() throws VelocityModelException {
+        double pVel = getVelocityLayer(0).getTopPVelocity();
+        double sVel = getVelocityLayer(0).getTopSVelocity();
+        double rho = getVelocityLayer(0).getTopDensity();
+        return calcReflTransCoef(0, 0, 0, pVel, sVel, rho, false);
+    }
+
     public ReflTrans calcReflTransCoef(double depth, boolean downgoing) throws VelocityModelException {
         if (!isDisconDepth(depth)) {
             throw new VelocityModelException("depth " + depth + " is not a discontinuity in the model");
         }
-        double abovePVel = evaluateAbove(depth, P_WAVE_CHAR);
-        double aboveSVel = evaluateAbove(depth, S_WAVE_CHAR);
-        double aboveRho = evaluateAbove(depth, DENSITY_CHAR);
-        double belowPVel = evaluateBelow(depth, P_WAVE_CHAR);
-        double belowSVel = evaluateBelow(depth, S_WAVE_CHAR);
-        double belowRho = evaluateBelow(depth, DENSITY_CHAR);
-        if (depth == 0) {
-            abovePVel = 0.0;
-            aboveSVel = 0.0;
-            aboveRho = 0.0;
+        double abovePVel = 0;
+        double aboveSVel = 0;
+        double aboveRho = 0;
+        double belowPVel = evaluateBelow(depth, VelocityModelMaterial.P_VELOCITY);
+        double belowSVel = evaluateBelow(depth, VelocityModelMaterial.S_VELOCITY);
+        double belowRho = evaluateBelow(depth, VelocityModelMaterial.DENSITY);
+        if (depth != 0) {
+            abovePVel = evaluateAbove(depth, VelocityModelMaterial.P_VELOCITY);
+            aboveSVel = evaluateAbove(depth, VelocityModelMaterial.S_VELOCITY);
+            aboveRho = evaluateAbove(depth, VelocityModelMaterial.DENSITY);
+        } else {
+            abovePVel = belowPVel;
         }
         return calcReflTransCoef(abovePVel, aboveSVel, aboveRho, belowPVel, belowSVel, belowRho, downgoing);
     }
@@ -1629,20 +1610,13 @@ public class VelocityModel implements Cloneable, Serializable {
             if (downgoing) {
                 throw new VelocityModelException("Downgoing to free surface is not possible");
             }
-            if (belowSVel == 0.0) {
-                rtCoef = new ReflTransFluidFreeSurface(belowPVel, belowRho);
-            } else {
-                rtCoef = new ReflTransFreeSurface(belowPVel, belowSVel, belowRho);
-            }
+            rtCoef = ReflTransFreeSurface.createReflTransFreeSurface(belowPVel, belowSVel, belowRho);
         } else if (belowPVel == 0.0 || belowRho == 0.0) {
             if (!downgoing) {
                 throw new VelocityModelException("Upgoing to inverted free surface is not possible");
             }
-            if (aboveSVel == 0.0) {
-                rtCoef = new ReflTransFluidFreeSurface(abovePVel, aboveRho);
-            } else {
-                rtCoef = new ReflTransFreeSurface(abovePVel, aboveSVel, aboveRho);
-            }
+            rtCoef = ReflTransFreeSurface.createReflTransFreeSurface(abovePVel, aboveSVel, aboveRho);
+
         } else if (aboveSVel == 0.0 && belowSVel == 0.0) {
             rtCoef = new ReflTransFluidFluid(abovePVel, aboveRho, belowPVel, belowRho);
             if (!downgoing) {
@@ -1671,7 +1645,4 @@ public class VelocityModel implements Cloneable, Serializable {
         return rtCoef;
     }
 
-    public static final char P_WAVE_CHAR = 'P';
-    public static final char S_WAVE_CHAR = 'S';
-    public static final char DENSITY_CHAR = 'D';
 }
